##
##  OpenPKG bf-db.pl - OpenPKG build farm database (used as shell tool)
##  Copyright (c) 2002-2005 The OpenPKG Project <http://www.openpkg.org/>
##  Copyright (c) 2002-2005 Ralf S. Engelschall <rse@engelschall.com>
##  Copyright (c) 2002-2005 Cable & Wireless <http://www.cw.com/>
##
##  Permission to use, copy, modify, and distribute this software for
##  any purpose with or without fee is hereby granted, provided that
##  the above copyright notice and this permission notice appear in all
##  copies.
##
##  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
##  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
##  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
##  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
##  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
##  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
##  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
##  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
##  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
##  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
##  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
##  SUCH DAMAGE.
##

require 5.008;
use strict;
use OpenPKG::Ctx;
use Data::Dumper;
use IO;
use IO::File;
use Getopt::Long;
use OSSP::cfg;
use DBI;
use DBD::SQLite;

#   configure optional debugging
$Data::Dumper::Purity = 1;
$Data::Dumper::Indent = 1;
$Data::Dumper::Terse  = 1;

#   determine OpenPKG information
my $ctx = new OpenPKG::Ctx;
my $progname = $ctx->cmdprog() . " " . $ctx->cmdname();
my $prefix = $ctx->prefix();

#   catch and hold time
my $time = time();

#   options with defaults
my $opt_config     = "$prefix/etc/openpkg-tools/bf.conf";
my $opt_help       = 0;
my $opt_maxretries = 5;
my $opt_verbose    = 0;
my $opt_color      = undef;
my $opt_inst       = undef;
my $opt_package    = undef;

sub verbose () {
    my ($msg) = @_;
    printf STDERR "$progname:DEBUG: %s\n", $msg if($opt_verbose);
}

sub usage ()
{
    my ($level) = @_;
    print "$progname:$level: Usage: $progname [common options][[specific options] command]\n" .
          "Available options:\n" .
          " -c|color=s         color (red|green)\n" .
          " -i|inst=s          instance (<host>-<arch>-<os>-<tag>)\n" .
          " -p|package=s       package (<name>-<version>-<release>)\n" .
          " -H|help            print out this usage page\n" .
          " -V|verbose         enable verbose run-time mode\n";
          " -C|config=s        specifiy config file ($opt_config)\n" .
          "Available commands:\n" .
          " begin              log beginning of build for instance and package\n" .
          " end                log end and color of build for instance and package\n";
    exit(1) if (defined $level);
    exit(0);
}

#   command line parsing
#
#   reserved for splitted -i|inst    => -h|host, -a|arch, -o|os, -t|tag
#   reserved for splitted -i|package => -n|name, -v|version, -r|release
Getopt::Long::Configure("bundling");
my $result = GetOptions(
    'c|color=s'      => \$opt_color,
    'i|inst=s'       => \$opt_inst,
    'p|package=s'    => \$opt_package,
    'C|config=s'     => \$opt_config,
    'H|help'         => \$opt_help,
    'V|verbose'      => \$opt_verbose
) || &usage("ERROR");
&usage() if ($opt_help);
( $#ARGV == 0 ) || &usage();

#   handle config
my $cfg = &readconfig($opt_config);
my $dbfile      = $cfg->{"master"}->{"dbfile"}      || die "mandatory setting master.dbfile undefined";
my $lockretries = $cfg->{"master"}->{"lockretries"} || 1;

my $dbh = &dbinit($dbfile);
if    ( $ARGV[0] =~ m|^init$|  ) { }
elsif ( $ARGV[0] =~ m|^begin$| ) { &status_begin($opt_inst, $opt_package); }
elsif ( $ARGV[0] =~ m|^end$|   ) { &status_end  ($opt_inst, $opt_package, $opt_color); }
else { &usage(); }
$dbh->disconnect();
undef $dbh;
exit(0);

#   this code reading config is shared between bf-mk.pl, bf-db.pl and
#   bf-ui.pl and a candidate to be moved into a perl module shared
#   between the three as soon as such a module is created.
#
sub subst ($$)
{
    my ($cfg, $txt) = @_;
    while ($txt =~ m|%{([^}]+)\.([^}]+)}|) {
        my ($section, $k) = ($1, $2);
        my $v = $cfg->{$section}->{$k};
        if (defined $v) {
            $txt =~ s|%{$section.$k}|$v|g;
        }
        else {
            die "undefined variable %{$section.$k}";
        }
    }
    return $txt;
}

sub readconfig($)
{
    my ($cfgfile) = @_;
    my $cfg = {};
    my $osspcfg = new OSSP::cfg::simple;

    die "no config file specified" unless (defined $cfgfile);
    my $io = new IO::File "<$cfgfile";
    die "cannot open \"$cfgfile\" for reading ($!)" if(not defined $io);
    my $txt = ''; $txt .= $_ while (<$io>);
    $io->close();

    $osspcfg->parse($txt) || die "error parsing config file";
    my $tree = $osspcfg->unpack(-index => '.*') || die "error unpacking config file";

    foreach my $section (qw(master shared driver slaves)) {
        foreach my $k (keys %{$tree->{$section}->[1]}) {
            my $v = $tree->{$section}->[1]->{$k}->[1];
            $v =~ s|%{([^}\.]+)}|%{$section.\1}|g; #expand shortcuts with section omitted
            $cfg->{$section}->{$k} = $v;
            &verbose("section=$section, k=$k, v=$v");
        }
    }
    foreach my $section (keys %{$cfg}) {
        foreach my $k (keys %{$cfg->{$section}}) {
            $cfg->{$section}->{$k} = &subst($cfg, $cfg->{$section}->{$k});
        }
    }
    return($cfg);
}

sub status_begin ($$)
{
    my ($inst, $package) = @_;
    my ($host, $arch, $os, $tag) = &splitinst($inst);
    my ($name, $version, $release) = &splitpackage($package);
    my ($sth, $rv, $retries);
    printf STDERR "DEBUG:$0: status_begin(%s, %s) host=%s, arch=%s, os=%s, tag=%s, name=%s, version=%s, release=%s;\n", $inst, $package, $host, $arch, $os, $tag, $name, $version, $release if ($opt_verbose);

    $sth = $dbh->prepare(
        "INSERT INTO status ( " .
        "    st_inst_host, st_inst_arch, st_inst_os, st_inst_tag, " .
        "    st_package_name, st_package_version, st_package_release, " .
        "    st_build_begin ) " .
        "VALUES ( ?, ?, ?, ?, ?, ?, ?, ? );"
    );
    $retries = $opt_maxretries; do {
        $rv = $sth->execute(
            $host, $arch, $os, $tag,
            $name, $version, $release,
            $time
            );
    } while (not $rv and $dbh->errstr =~ m|lock|i and sleep(1) and $retries-- >= 1);
    printf STDERR "DEBUG:$0: countdown from %d to %d\n", $opt_maxretries, $retries if($opt_verbose);
    if (not $rv) {
        print STDERR "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
        exit(2);
	}

    $sth = $dbh->prepare(
        "INSERT INTO processing ( " .
        "    pr_inst_host, pr_inst_arch, pr_inst_os, pr_inst_tag, " .
        "    pr_package_name, pr_package_version, pr_package_release, " .
        "    pr_heartbeat ) " .
        "VALUES ( ?, ?, ?, ?, ?, ?, ?, ? );"
    );
    $rv = $sth->execute(
        $host, $arch, $os, $tag,
        $name, $version, $release,
        $time
        );
    if (not $rv) {
        print STDERR "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
        exit(2);
    }
}

sub status_end ($$$)
{
    my ($inst, $package, $color) = @_;
    my ($host, $arch, $os, $tag) = &splitinst($inst);
    my ($name, $version, $release) = &splitpackage($package);
    my ($sth, $rv, $retries);
    printf STDERR "DEBUG:$0: status_end(%s, %s, %s) host=%s, arch=%s, os=%s, tag=%s, name=%s, version=%s, release=%s;\n", $inst, $package, $color, $host, $arch, $os, $tag, $name, $version, $release if ($opt_verbose);

    #   special case to handle UPDATE from nonexisting row
    #   this can happen if the database was erased between build begin and build end
    #   in this particular case we have a end timestamp but no begin timestamp
    #
    {
        $sth = $dbh->prepare(
            "SELECT st_package_name FROM status WHERE ( " .
            "    st_inst_host = ? and st_inst_arch = ? and st_inst_os = ? and st_inst_tag = ? and " .
            "    st_package_name = ? );"
        );
        $rv = $sth->execute(
            $host, $arch, $os, $tag,
            $name
        );
        if (not $rv) {
            print STDERR "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
            exit(2);
        }
        $rv = $sth->fetchrow_array;
        if ($sth->rows < 1) {
            $sth = $dbh->prepare(
                "INSERT INTO status ( " .
                "    st_inst_host, st_inst_arch, st_inst_os, st_inst_tag, " .
                "    st_package_name, st_package_version, st_package_release ) " .
                "VALUES ( ?, ?, ?, ?, ?, ?, ? );"
            );
            my $rv = $sth->execute(
                $host, $arch, $os, $tag,
                $name, $version, $release,
                );
            if (not $rv) {
                print STDERR "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
                exit(2);
            }
        }
    }

    if (defined $color) {
        $sth = $dbh->prepare(
            "UPDATE status SET st_build_end = ?, st_build_color = ? WHERE ( " .
            "    st_inst_host = ? and st_inst_arch = ? and st_inst_os = ? and st_inst_tag = ? and " .
            "    st_package_name = ? );"
        );
        $retries = $opt_maxretries; do {
            $rv = $sth->execute(
                $time, $color,
                $host, $arch, $os, $tag,
                $name
            );
        } while (not $rv and $dbh->errstr =~ m|lock|i and sleep(1) and $retries-- >= 1);
        printf STDERR "DEBUG:$0: countdown from %d to %d\n", $opt_maxretries, $retries if($opt_verbose);
        if (not $rv) {
            print STDERR "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
            exit(2);
        }
    }
    else {
        $sth = $dbh->prepare(
            "UPDATE status SET st_build_end = ? WHERE ( " .
            "    st_inst_host = ? and st_inst_arch = ? and st_inst_os = ? and st_inst_tag = ? and " .
            "    st_package_name = ? );"
        );
        $rv = $sth->execute(
            $time,
            $host, $arch, $os, $tag,
            $name
        );
    }
    if (not $rv) {
        print STDERR "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
        exit(2);
    }

    $sth = $dbh->prepare(
        "INSERT INTO processing ( " .
        "    pr_inst_host, pr_inst_arch, pr_inst_os, pr_inst_tag, " .
        "    pr_package_name, pr_package_version, pr_package_release, " .
        "    pr_heartbeat ) " .
        "VALUES ( ?, ?, ?, ?, ?, ?, ?, ? );"
    );
    $retries = $opt_maxretries; do {
        $rv = $sth->execute(
            $host, $arch, $os, $tag,
            "", "", "",
            $time
            );
    } while (not $rv and $dbh->errstr =~ m|lock|i and sleep(1) and $retries-- >= 1);
    printf STDERR "DEBUG:$0: countdown from %d to %d\n", $opt_maxretries, $retries if($opt_verbose);
    if (not $rv) {
        print STDERR "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
        exit(2);
    }

}

#   open/create database and create all missing tables on the fly
sub dbinit ($)
{
    my ($dbfile) = @_;
    my $dbh = DBI->connect("dbi:SQLite:dbname=$dbfile", "", "", { RaiseError => 0, AutoCommit => 1 });
    if (not defined $dbh) {
        print STDERR "ERROR:$0: \"cannot open database file \"$dbfile\"\n";
        exit(2);
    }

	my $tables = $dbh->selectall_hashref("SELECT name FROM sqlite_master WHERE type='table' ORDER BY name;", "name");

	unless ($tables->{status}) {
        printf STDERR "DEBUG:$0: creating \"status\" table in database file \"%s\"\n", $dbfile if ($opt_verbose);
        my $sql = '';
        $sql .= "CREATE TABLE status";                         # field description                  | example
		$sql .= "(\n";                                         # -----------------------------------+-----------
        $sql .= "    st_inst_host       TEXT NOT NULL,\n";     # 1st p. PLATFORM in Makefile.rules  | bsd1
        $sql .= "    st_inst_arch       TEXT NOT NULL,\n";     # 2st p. PLATFORM in Makefile.rules  | ix86
        $sql .= "    st_inst_os         TEXT NOT NULL,\n";     # 3rd p. PLATFORM in Makefile.rules  | freebsd5.3
        $sql .= "    st_inst_tag        TEXT NOT NULL,\n";     # 4th p. PLATFORM in Makefile.rules  | openpkg
        $sql .= "    st_package_name    TEXT NOT NULL,\n";     # 1st part of $pkg in Makefile.rules | a2ps
        $sql .= "    st_package_version TEXT NOT NULL,\n";     # 2nd part of $pkg in Makefile.rules | 4.13b
        $sql .= "    st_package_release TEXT NOT NULL,\n";     # 3rd part of $pkg in Makefile.rules | 20040816
        $sql .= "    st_build_begin     INTEGER DEFAULT 0,\n"; # time_t build began, 0 if unknown   | 1105654183
        $sql .= "    st_build_end       INTEGER DEFAULT 0,\n"; # time_t build ended, 0 if unknown   | 1105654224
        $sql .= "    st_build_color     TEXT DEFAULT \"\",\n"; # red/green depending on binary RPM  | green
        $sql .= "PRIMARY KEY ( st_inst_host, st_inst_arch, st_inst_os, st_inst_tag, st_package_name ) ON CONFLICT REPLACE\n";
        $sql .= ");\n";
        $dbh->do($sql);
    };

	unless ($tables->{processing}) {
        printf STDERR "DEBUG:$0: creating \"processing\" table in database file \"%s\"\n", $dbfile if ($opt_verbose);
        my $sql = '';
        $sql .= "CREATE TABLE processing";                     # field description                  | example
		$sql .= "(\n";                                         # -----------------------------------+-----------
        $sql .= "    pr_inst_host       TEXT NOT NULL,\n";     # 1st p. PLATFORM in Makefile.rules  | bsd1
        $sql .= "    pr_inst_arch       TEXT NOT NULL,\n";     # 2st p. PLATFORM in Makefile.rules  | ix86
        $sql .= "    pr_inst_os         TEXT NOT NULL,\n";     # 3rd p. PLATFORM in Makefile.rules  | freebsd5.3
        $sql .= "    pr_inst_tag        TEXT NOT NULL,\n";     # 4th p. PLATFORM in Makefile.rules  | openpkg
        $sql .= "    pr_package_name    TEXT,\n";              # 1st part of $pkg in Makefile.rules | a2ps
        $sql .= "    pr_package_version TEXT,\n";              # 2nd part of $pkg in Makefile.rules | 4.13b
        $sql .= "    pr_package_release TEXT,\n";              # 3rd part of $pkg in Makefile.rules | 20040816
        $sql .= "    pr_heartbeat       INTEGER DEFAULT 0,\n"; # time_t last update, 0 if unknown   | 1105654183
        $sql .= "PRIMARY KEY ( pr_inst_host, pr_inst_arch, pr_inst_os, pr_inst_tag ) ON CONFLICT REPLACE\n";
        $sql .= ");\n";
        $dbh->do($sql);
    };

	unless ($tables->{class}) {
        printf STDERR "DEBUG:$0: creating \"class\" table in database file \"%s\"\n", $dbfile if ($opt_verbose);
        my $sql = '';
        $sql .= "CREATE TABLE class";                          # field description                  | example
		$sql .= "(\n";                                         # -----------------------------------+-----------
        $sql .= "    cl_package_name    TEXT NOT NULL,\n";     # 1st part of $pkg in Makefile.rules | a2ps
        $sql .= "    cl_package_class   TEXT,\n";              # CORE|BASE|PLUS|EVAL|JUNK           | CORE
        $sql .= "PRIMARY KEY ( cl_package_name ) ON CONFLICT REPLACE\n";
        $sql .= ");\n";
        $dbh->do($sql);
    };

    return $dbh;
}

#   split inst "bsd1-ix86-freebsd5.3-openpkg" into host="bsd1", arch="ix86", os="freebsd5.3" and tag="openpkg"
sub splitinst ($)
{
    my ($inst) = @_;
    my ($host, $arch, $os, $tag);
    if ($inst =~ m|^(.+)-([^-]+)-([^-]+)-([^-]+)$|) {
        ($host, $arch, $os, $tag) = ($1, $2, $3, $4);
    };
    return ($host, $arch, $os, $tag);
}

#   split package "perl-openpkg-5.8.6-20041203" into name="perl-openpkg", version="5.8.6" and release="20041203"
sub splitpackage ($)
{
    my ($package) = @_;
    my ($name, $version, $release);
    if ($package =~ m|^(.+)-([^-]+)-([^-]+)$|) {
        ($name, $version, $release) = ($1, $2, $3);
    };
    return ($name, $version, $release);
}

#   die gracefully ;-)
exit(0);

__END__

=pod

=head1 NAME

B<bfdb> - OpenPKG build farm database

=head1 SYNOPSIS

Shell tool to feed build farm database.

  begin -i <host>-<arch>-<os>-<tag> \
        -p <name>-<version>-<release>

  end   -i <host>-<arch>-<os>-<tag> \
        -p <name>-<version>-<release> \
        -c <color>

=head1 DESCRIPTION

    begin

    logs the beginning of a package build.

      -i|inst=s          instance (<host>-<arch>-<os>-<tag>)
      -p|package=s       package (<name>-<version>-<release>)

    end

      logs the end of a package build and
      the color (result).

      -i|inst=s          instance (<host>-<arch>-<os>-<tag>)
      -p|package=s       package (<name>-<version>-<release>)
      -c|color=s         color (red|green)

  common options

      [-C|config=s]      specifiy config file
      [-H|help]          print out this usage page
      [-V|verbose]       enable verbose run-time mode

=cut

