##
##  OpenPKG bf-mk.pl - OpenPKG build farm maker
##  Copyright (c) 2002-2005 The OpenPKG Project <http://www.openpkg.org/>
##  Copyright (c) 2002-2005 Ralf S. Engelschall <rse@engelschall.com>
##  Copyright (c) 2002-2005 Cable & Wireless <http://www.cw.com/>
##
##  Permission to use, copy, modify, and distribute this software for
##  any purpose with or without fee is hereby granted, provided that
##  the above copyright notice and this permission notice appear in all
##  copies.
##
##  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
##  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
##  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
##  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
##  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
##  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
##  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
##  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
##  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
##  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
##  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
##  SUCH DAMAGE.
##

require 5.008;
use strict;
use OpenPKG::Ctx;
use Data::Dumper;
use IO;
use IO::File;
use Getopt::Long;
use OSSP::cfg;

#   configure optional debugging
$Data::Dumper::Purity = 1;
$Data::Dumper::Indent = 1;
$Data::Dumper::Terse  = 1;

#   determine OpenPKG information
my $ctx = new OpenPKG::Ctx;
my $progname = $ctx->cmdprog() . " " . $ctx->cmdname();
my $prefix = $ctx->prefix();

#   options with defaults
my $opt_config     = "$prefix/etc/openpkg-tools/bf.conf";
my $opt_outfile    = undef;
my $opt_bashrc     = 0;
my $opt_bashlogin  = 0;
my $opt_key        = 0;
my $opt_master     = 0;
my $opt_slave      = 0;
my $opt_oneshot    = 0;
my $opt_loop       = 0;
my $opt_init       = 0;
my $opt_bfconf     = 0;
my $opt_uicgi      = 0;
my $opt_query      = undef;
my $opt_help       = 0;
my $opt_verbose    = 0;

#   internal handling of fatal errors
BEGIN {
    $SIG{__DIE__} = sub {
        my ($msg) = @_;
        printf "$progname:ERROR: %s\n", $msg;
        exit(1);
    };
}

sub shtoolechoe () {
    my ($msg) = @_;
    my $Bon  = chr(0x1b) . chr(0x5b) . chr(0x31) . chr(0x6d); #bold on 
    my $Boff = chr(0x1b) . chr(0x5b) . chr(0x6d);             #bold off
    $msg =~ s|%B|$Bon|g;
    $msg =~ s|%b|$Boff|g;
    printf "%s\n", $msg;
}

sub verbose () {
    my ($msg) = @_;
    printf STDERR "$progname:DEBUG: %s\n", $msg if($opt_verbose);
}

sub usage ()
{
    my ($level) = @_;
    $level .= ":" if (defined $level);
    print "$progname:$level Usage: $progname [options]\n" .
          "Available options:\n" .
          " -o|outfile=s       write output to file\n" .
          " -r|bashrc          generate .bashrc\n" .
          " -p|bashlogin       generate .bash_login profile\n" .
          " -k|key             generate .ssh/ keys\n" .
          " -m|master          generate bin/master wrapper\n" .
          " -s|slave           generate bin/slave script\n" .
          " -1|oneshot         run master once\n" .
          " -l|loop            run master in a loop\n" .
          " -i|init            init mode\n" .
          " -b|bfconf          initialize bf.conf\n" .
          " -u|uicgi           initialize bf-ui.cgi\n" .
          " -q|query=s         query variable from bf.conf\n" .
          " -C|config=s        specifiy config file ($opt_config)\n" .
          " -H|help            print out this usage page\n" .
          " -V|verbose         enable verbose run-time mode\n" .
          "Notes on init mode";
          " with -i the master gets initialized\n";
          " with -[bu] only selected files are initialized, otherwise all\n" .
          "Notes on master mode";
          " with -[1l] the application executes the MASTER role\n" .
          "Notes on query mode";
          " with -q variables can be printed, e.g. \"prefix=%{master.prefix}\"\n";
          "Notes on shared setup mode";
          " with -[rlkms] only selected files are generated, otherwise all\n" .
          " with -o is written to given file otherwise to proper destination\n";
    exit(1) if (defined $level);
    exit(0);
}

#   command line parsing
Getopt::Long::Configure("bundling");
my $result = GetOptions(
    'o|outfile=s'    => \$opt_outfile,
    'r|bashrc'       => \$opt_bashrc,
    'p|bashlogin'    => \$opt_bashlogin,
    'k|key'          => \$opt_key,
    'm|master'       => \$opt_master,
    's|slave'        => \$opt_slave,
    '1|oneshot'      => \$opt_oneshot,
    'l|loop'         => \$opt_loop,
    'i|init'         => \$opt_init,
    'b|bfconf'       => \$opt_bfconf,
    'u|uicgi'        => \$opt_uicgi,
    'q|query=s'      => \$opt_query,
    'C|config=s'     => \$opt_config,
    'H|help'         => \$opt_help,
    'V|verbose'      => \$opt_verbose
) || &usage("ERROR");
&usage() if ($opt_help);

if ($opt_outfile and ($opt_bashrc + $opt_bashlogin + $opt_key + $opt_master + $opt_slave != 1)) {
    die "outfile specification requires a single file to be generated";
}

#   main switch
if ($opt_init) {
    ( -d $prefix ) || die "internal error - prefix undefined or not a directory";

    #   if no files selected default to all files selected
    unless ($opt_bfconf or $opt_uicgi) {
        $opt_bfconf = 1;
        $opt_uicgi = 1;
    }

    &shtoolechoe("%B++ MASTER:$prefix: init%b");
    &createmasterstructure($prefix);
    &writefile("$prefix/etc/openpkg-tools/bf.conf",   $opt_outfile, 0644, &init_bfconf()) if($opt_bfconf);
    &writefile("$prefix/cgi/openpkg-tools/bf-ui.cgi", $opt_outfile, 0755, &init_uicgi())  if($opt_uicgi);
}
else {
    my $cfg = &readconfig($opt_config);
    my $shared = $cfg->{"shared"}->{"prefix"} || die "mandatory setting shared.prefix undefined";

    #   if no files selected default to all files selected
    unless ($opt_bashrc or $opt_bashlogin or $opt_key or $opt_master or $opt_slave) {
        $opt_bashrc = 1;
        $opt_bashlogin = 1;
        $opt_key = 1;
        $opt_master = 1;
        $opt_slave = 1;
    }

    if ($opt_oneshot) {
        &shtoolechoe("%B++ MASTER:$shared: oneshot%b");
        &oneshot($cfg);
    }
    elsif ($opt_loop) {
        my $round = 1;
        while(1) {
            &shtoolechoe("%B++ MASTER:$shared: loop%b round $round");
            &oneshot($cfg);
            $round++;
        };
    }
    elsif ($opt_query) {
        &shtoolechoe("%B++ MASTER:$shared: query%b");
        printf "%s\n", &subst($cfg, $opt_query);
    }
    else {
        &shtoolechoe("%B++ MASTER:$shared: setup%b");
        &createsharedstructure($shared);
        &gen_key($cfg) if($opt_key);
        &writefile("$shared/.bashrc",     $opt_outfile, 0644, &gen_bashrc($cfg))    if($opt_bashrc);
        &writefile("$shared/.bash_login", $opt_outfile, 0644, &gen_bashlogin($cfg)) if($opt_bashrc);
        &writefile("$shared/bin/master",  $opt_outfile, 0755, &gen_master($cfg))    if($opt_master);
        &writefile("$shared/bin/slave",   $opt_outfile, 0755, &gen_slave($cfg))     if($opt_slave);
    }
}
#   die gracefully ;-)
exit(0);

#   this code reading config is shared between bf-mk.pl, bf-db.pl and
#   bf-ui.pl and a candidate to be moved into a perl module shared
#   between the three as soon as such a module is created.
#
sub subst ($$)
{
    my ($cfg, $txt) = @_;
    while ($txt =~ m|%{([^}]+)\.([^}]+)}|) {
        my ($section, $k) = ($1, $2);
        my $v = $cfg->{$section}->{$k};
        if (defined $v) {
            $txt =~ s|%{$section.$k}|$v|g;
        }
        else {
            die "undefined variable %{$section.$k}";
        }
    }
    return $txt;
}

sub readconfig($)
{
    my ($cfgfile) = @_;
    my $cfg = {};
    my $osspcfg = new OSSP::cfg::simple;

    die "no config file specified" unless (defined $cfgfile);
    my $io = new IO::File "<$cfgfile";
    die "cannot open \"$cfgfile\" for reading ($!)" if(not defined $io);
    my $txt = ''; $txt .= $_ while (<$io>);
    $io->close();

    $osspcfg->parse($txt) || die "error parsing config file";
    my $tree = $osspcfg->unpack(-index => '.*') || die "error unpacking config file";

    foreach my $section (qw(master shared driver slaves)) {
        foreach my $k (keys %{$tree->{$section}->[1]}) {
            my $v = $tree->{$section}->[1]->{$k}->[1];
            $v =~ s|%{([^}\.]+)}|%{$section.\1}|g; #expand shortcuts with section omitted
            $cfg->{$section}->{$k} = $v;
            &verbose("section=$section, k=$k, v=$v");
        }
    }
    foreach my $section (keys %{$cfg}) {
        foreach my $k (keys %{$cfg->{$section}}) {
            $cfg->{$section}->{$k} = &subst($cfg, $cfg->{$section}->{$k});
        }
    }
    return($cfg);
}

sub writefile($$$$)
{
    my ($testfile, $outfile, $perms, $data) = @_;
    $outfile = $testfile if (not $outfile);
    if ($outfile eq "-") {
        print "$data";
    }
    else {
        my $io = new IO::File "$outfile", O_RDWR|O_CREAT|O_TRUNC, $perms;
        die "cannot open \"$outfile\" for creation/writing ($!)" if(not defined $io);
        &verbose("writing \"$outfile\"");
        print $io "$data";
        $io->close();
    }
}

sub createsharedstructure($)
{
    my ($shared) = @_;
    foreach my $dir (qw(bin ctl dst log src src/master src/master.norestriction tmp tmp/master)) {
        my $target = $shared . "/" . $dir;
        &verbose("creating $target/");
        if (not -d $target) {
            mkdir $target, 0775 || die "cannot make directory \"$target\"";
        }
    }
}

sub createmasterstructure($)
{
    my ($prefix) = @_;
    foreach my $dir (qw(etc var cgi)) {
        my $target = $prefix . "/" . $dir . "/openpkg-tools";
        &verbose("creating $target/");
        if (not -d $target) {
            mkdir $target, 0775 || die "cannot make directory \"$target\"";
        }
    }
}

sub oneshot ($)
{
    my ($cfg) = @_;

    #   handle config
    my $shared     = $cfg->{"shared"}->{"prefix"}   || die "mandatory setting shared.prefix undefined";
    my $slaves     = $cfg->{"slaves"}->{"prefix"}   || die "mandatory setting slaves.prefix undefined";
    my $classes    = $cfg->{"master"}->{"classes"}  || die "mandatory setting master.classes undefined";
    my $rsync      = $cfg->{"master"}->{"rsync"}    || die "mandatory setting master.rsync undefined";
    my $src2make   = $cfg->{"master"}->{"src2make"} || die "mandatory setting master.src2make undefined";
    my $source     = $cfg->{"master"}->{"source"}   || die "mandatory setting master.source undefined";
    my $nosource   = $cfg->{"master"}->{"nosource"};
    my $skippkg    = $cfg->{"master"}->{"skippkg"};

    #   prepare directory structure
    &shtoolechoe("%B++ MASTER:$shared: prepare directory structure%b");
    &createsharedstructure($shared);

    #   download sources and remove those to be skipped
    &shtoolechoe("%B++ MASTER:$shared: download sources and remove those to be skipped%b");
    system("$rsync -rt --delete --exclude \"Makefile*\" \"$source\" \"$shared/src/master/\"") == 0 || die("cannot rsync source RPMs from \"$source\" ($!)");
    foreach my $p (split(/ /, $skippkg)) {
        #FIXME obey name-version-release instead only unlinking
        for my $f (<$shared/src/master/$p>) {
            unlink $f || warn "cannot remove \"$shared/src/master/$p\" ($!)";
        }
    }

    #   download sources with norestriction to replace .nosrc.rpm
    &shtoolechoe("%B++ MASTER:$shared: download sources with norestriction to replace .nosrc.rpm%b");
    if ($nosource) {
        system("$rsync -rt --delete --exclude \"Makefile*\" \"$nosource\" \"$shared/src/master.norestriction/\"") == 0 || die("cannot rsync nosource RPMs from \"$nosource\" ($!)");
    }
    foreach my $package (<$shared/src/master/*.nosrc.rpm>) {
        unlink $package || die "cannot erase \"$package\"";
        $package =~ s|^.*/||;
        $package =~ s|\.nosrc\.rpm$||;
        my $source="$shared/src/master.norestriction/$package.src.rpm";
        my $target="$shared/src/$package.src.rpm";
        if (-f $source) {
            link $source, $target || die "cannot link \"$source\" to \"$target\"";
        }
        else {
            warn "package with norestriction for $source not available";
        }
    }

    #   run src2make
    &shtoolechoe("%B++ MASTER:$shared: run src2make%b");
    system("$src2make --prefix=\"$slaves\" " .
                     "--srcdir=\"$shared/src/master\" " .
                     "--outdir=\"$shared/src/master\" " .
                     "--tmpdir=\"$shared/tmp/master\" " .
                     "--classes=\"$classes\" " .
                     "--verbose") == 0 || die "src2make failed ($!)";
}

sub gen_key ($)
{
    my ($cfg) = @_;

    #   handle config
    my $shared     = $cfg->{"shared"}->{"prefix"}    || die "mandatory setting shared.prefix undefined";
    my $sshkeygen  = $cfg->{"master"}->{"sshkeygen"} || die "mandatory setting master.sshkeygen undefined";

    #   generate server DSA (SSH2) key
    my $key = "$shared/.ssh/id_dsa";
    if ( ! -f $key or ! -s $key ) {
        system("$sshkeygen -t dsa -b 2048 " .
               "-f $key " .
               "-N '' -C \"(WORKER:$shared)\" 1>&2") == 0 || die "ssh-keygen failed to create \"$key\" ($!)";
    }

    #   generate authorized_keys
    my $auth = "$shared/.ssh/authorized_keys";
    if ( ! -f $auth or ! -s $auth ) {
        link $key.".pub", $auth || die "creation of \"$auth\" failed ($!)";
    }
}

sub gen_bashrc ($)
{
    my ($cfg) = @_;
    my $txt = <<'EOT';
if [ ".$HOME" != ".%{shared.prefix}" ]; then
    echo ".bashrc:ERROR: HOME does not match %{shared.prefix}"
fi
export PATH=%{shared.prefix}/bin:$PATH
export PS1="worker@`%{driver.hostname} | %{driver.sed} -e 's;\..*$;;'`\$ "
unset TMOUT
EOT
    return &subst($cfg, $txt);
}

sub gen_bashlogin ($)
{
    my ($cfg) = @_;
    my $txt = <<'EOT';
. .bashrc
EOT
    return &subst($cfg, $txt);
}

sub gen_master ($)
{
    my ($cfg) = @_;
    my $txt = <<'EOT';
#!%{driver.sh}
##
##  This file was created by "openpkg bf-mk" and will be overwritten on
##  any configuration change. Manual edits are a waste of time.
##

exec %{master.bfmk} "$@"
EOT
    return &subst($cfg, $txt);
}

sub gen_slave ($)
{
    my ($cfg) = @_;
    my $txt = <<'EOT';
#!%{driver.sh}
##
##  This file was created by "openpkg bf-mk" and will be overwritten on
##  any configuration change. Manual edits are a waste of time.
##

warn ()
{
    echo "$0:WARNING: $1"
}

die ()
{
    echo "$0:ERROR: $1"
    exit 1
}

assert ()
{
    [ ".%{shared.prefix}" != . ] || die "SHARED directory not configured"
    [ -d "%{shared.prefix}" ] || die "shared prefix \"%{shared.prefix}\" does not exist on this machine"
    cd %{shared.prefix} || die "cannot change directory to shared prefix \"%{shared.prefix}\""
    [ -x "%{driver.shtool}" ] || die "driver shtool(1) \"%{driver.shtool}\" not executable"
    [ -x "%{driver.sudo}" ] || die "driver sudo(1) \"%{driver.sudo}\" not executable"
    [ ".%{slaves.prefix}" != . ] || die "SLAVES prefix not configured"
    [ -d "%{slaves.prefix}" ] || die "slave prefix \"%{slaves.prefix}\" does not exist on this machine"
    [ -d "%{slaves.ltmp}" ] || die "large temporary disks space \"%{slaves.ltmp}\" missing"
    [ ".%{slaves.bfdb}" != . ] || die "SLAVES bfdb not configured"
}

oneshot ()
{
    #   creating SHARED directory structure for SLAVE
    %{driver.shtool} echo -e "%B++ SLAVE@$HOST:%{shared.prefix}: creating SHARED directory structure for SLAVE%b"
    for dir in ctl dst log src tmp; do
        target="%{shared.prefix}/$dir/$HOST-$ARCH-$OS"
        %{driver.shtool} mkdir -f -p -m 775 $target || die "cannot create directory \"$target\""
    done

    #   prepare SRPMs for SLAVE
    %{driver.shtool} echo -e "%B++ SLAVE@$HOST:%{shared.prefix}: prepare SRPMs for SLAVE%b"
    source="%{shared.prefix}/src/master"
    target="%{shared.prefix}/src/$HOST-$ARCH-$OS"
    %{driver.cp} $source/Makefile* $target/
    %{driver.rm} -f $target/*.src.rpm
    for srpm in `cd $source && ls -1 *.src.rpm | sort`; do
        %{driver.ln} $source/$srpm $target/
    done

    #   erase packages from SLAVE OpenPKG instance
    %{driver.shtool} echo -e "%B++ SLAVE@$HOST:%{shared.prefix}: erase packages from SLAVE OpenPKG instance%b"
    ( cd $target && %{driver.make} \
        LOG="%{shared.prefix}/log/$HOST-$ARCH-$OS" \
        SRC="%{shared.prefix}/src/$HOST-$ARCH-$OS" \
        DST="%{shared.prefix}/dst/$HOST-$ARCH-$OS" \
        TMP="%{shared.prefix}/tmp/$HOST-$ARCH-$OS" \
        EXEC_BFDB="%{slaves.bfdb}" \
        EXEC_ROOT="%{driver.sudo}" \
        EXEC_HOST="$HOST" \
        erase \
    )

    #   SLAVE building packages
    %{driver.shtool} echo -e "%B++ SLAVE@$HOST:%{shared.prefix}: SLAVE building packages%b"
    ( cd $target && %{driver.make} \
        LOG="%{shared.prefix}/log/$HOST-$ARCH-$OS" \
        SRC="%{shared.prefix}/src/$HOST-$ARCH-$OS" \
        DST="%{shared.prefix}/dst/$HOST-$ARCH-$OS" \
        TMP="%{shared.prefix}/tmp/$HOST-$ARCH-$OS" \
        EXEC_BFDB="%{slaves.bfdb}" \
        EXEC_ROOT="%{driver.sudo}" \
        EXEC_HOST="$HOST" \
        try-build-latest \
    )
}

wipe ()
{
    #   wipe out SHARED areas belonging to SLAVE
    %{driver.shtool} echo -e "%B++ SLAVE@$HOST:%{shared.prefix}: wipe out SHARED areas belonging to SLAVE%b"
    for dir in bin ctl log src tmp; do
        target="%{shared.prefix}/$dir/$HOST-$ARCH-$OS"
        %{driver.rm} -rf || warn "cannot remove directory \"$target\""
    done
}

usage ()
{
    level="$1";
    if [ ".$1" != . ]; then level="$level:"; fi

    ( echo "$0:$level Usage: $0 [options]"
      echo "Available options:"
      echo " -1|oneshot         run master once"
      echo " -l|loop            run master in a loop"
      echo " -H|help            print out this usage page"
      echo "Notes on slave mode"
      echo " with -[1l] the application executes the SLAVE role"
      echo "To reconfigure settings, change bf.conf on the MASTER"
      echo " and regenerate environment using \"master\""
    ) >&2
    if [ ".$1" != . ]; then exit 1; fi
    exit 0
}

loop ()
{
    round=1
    while [ 1 ]; do
        %{driver.shtool} echo -e "%B++ SLAVE@$HOST:%{shared.prefix}: loop%b round $round"
        oneshot
        round=`expr $round + 1`
    done
}

#main()
assert

#   configuration queried from SLAVE
HOST=`%{driver.hostname} | %{driver.sed} -e 's;\..*$;;'`
ARCH=`%{slaves.rpm} --eval '%{l_host_arch}'`
OS=`%{slaves.rpm} --eval '%{l_host_os}'`

cmd=""
if [ ".$1" = ".-1" -o ".$1" = ".--oneshot" ]; then cmd="oneshot"; fi
if [ ".$1" = ".-l" -o ".$1" = ".--loop" ];    then cmd="loop"; fi
if [ ".$1" = ".-H" -o ".$1" = ".--help" ];    then cmd="help"; fi
if [ ".$cmd" = . ]; then usage "ERROR"; fi
if [ ".$cmd" = .help ]; then usage; fi

if [ ".$cmd" = .oneshot ]; then oneshot; fi
if [ ".$cmd" = .loop ]; then loop; fi
exit 1

EOT
    return &subst($cfg, $txt);
}

sub init_bfconf ()
{
    my $txt = <<'EOT';
##
##  OpenPKG build farm configuration
##

master {
    #   OpenPKG instance on SLAVES doing the build
    #   openpkg-team using "/openpkg"
    #
    prefix      "@l_prefix@";
    openpkg     "%{prefix}/bin/openpkg";
    shtool      "%{openpkg} shtool";
    src2make    "%{openpkg} src2make";
    rpm         "%{openpkg} rpm";
    bfmk        "%{openpkg} bf-mk";
    bfui        "%{openpkg} bf-ui";
    bfdb        "%{openpkg} bf-db";
    db          "%{prefix}/var/openpkg-tools/bf/db.sqlite";
    rsync       "%{prefix}/bin/rsync";
    screen      "%{prefix}/bin/screen";
    shell       "%{prefix}/bin/bash";
    sshkeygen   "%{prefix}/bin/ssh-keygen";
    sshadd      "%{prefix}/bin/ssh-add";
    sshagent    "%{prefix}/bin/ssh-agent";
    host        "master.example.com";

    #   database file
    #   openpkg-team using "%{prefix}/var/openpkg-tools/bf/db.sqlite"
    dbfile      "%{prefix}/var/openpkg-tools/bf/db.sqlite";

    #   max. retries to get database lock
    #   openpkg-team using 5
    lockretries 5;

    #   location to download SRPMs from using rsync(1)
    #   openpkg-team using "rsync://rsync.openpkg.org/openpkg-pre"
    #
    source      "rsync://rsync.openpkg.org/openpkg-ftp/current/SRC/";

    #   location where master can download usable replacements for .nosrc.rpm packages
    #   Empty setting inhibits download and effectivly disables processing of .nosrc.rpm packages
    #   openpkg-team using "/e/openpkg/PKG/src/"
    #
    nosource    "rsync://rsync.openpkg.org/openpkg-ftp/private/SRC/";

    #   let master prepare Makefile(.rules) for packages of the following classes
    #   openpkg-team using "+CORE+BASE-PLUS-EVAL-JUNK" for final release build
    classes     "+CORE-BASE-PLUS-EVAL-JUNK";

    #   let master skip Makefile(.rules) preparation for packages matching the following shell wildcard patterns
    #   openpkg-team using ""
    skippkg     "";
};

shared {
    #   folder shared between MASTER and SLAVES, e.g. NFS mount
    #   MASTER writes SRPMs, Makefile and Makefile.rules to $SHARED/src/master/
    #   MASTER uses $SHARED/tmp/ for caching SRPMs spec files
    #   MASTER picks up logs from $SHARED/log/
    #   SLAVES write logs to $SHARED/log and binaries to $SHARED/bin/
    #   openpkg-team using "/e/openpkg/bf"
    #
    prefix      "@l_prefix@/var/openpkg-tools/bf/shared";

    #   User account launching the actual work
    #   Must exist on MASTER and all SLAVES (consider NIS)
    #   Must have valid bash login shell on MASTER and all SLAVES
    #   Must have $HOME set to $SHARED
    #   Must be able to ssh to MASTER and all SLAVES
    #   logs in to MASTER and launches/reconnects screen(1) via "make" aka "make all" aka "make worker"
    #   openpkg-team using "openpkg-bf"
    #
    worker      "worker";
};

slaves {
    #   OpenPKG instance on SLAVES doing the build
    #   openpkg-team using "/openpkg"
    #
    prefix      "/bf/slaves";
    openpkg     "%{prefix}/bin/openpkg";
    rpm         "%{openpkg} rpm";
    bfdb        "ssh %{shared.worker}@%{master.host} %{master.bfdb}";

    #   large and fast temporary disk space for building
    #   openpkg-team using "/ltmp/openpkg-bf"
    ltmp        "/var/tmp";
};


driver {
    #   Place to find common (development) tools used by SLAVES
    #   Might be a OpenPKG instance but can be OS stuff if available
    #   openpkg-team using "/usr/opkg"
    #
    prefix      "/bf/driver";
    gcc         "%{prefix}/bin/gcc";
    ld          "%{prefix}/bin/ld";
    make        "%{prefix}/bin/make";
    sh          "/bin/sh";
    shtool      "%{prefix}/lib/openpkg/shtool";
    sudo        "%{prefix}/bin/sudo";
    sudoers     "%{prefix}/etc/sudo/sudoers";
    hostname    "hostname";
    sed         "sed";
    cp          "cp";
    ln          "ln";
    rm          "rm";
};

EOT
    $txt =~ s/\@l_prefix\@/$prefix/g;
    return $txt;
}

sub init_uicgi ($)
{
    my ($cfg) = @_;
    my $txt = <<'EOT';
#!@l_prefix@/lib/openpkg/bash
exec @l_prefix@/bin/openpkg bf-ui
EOT
    $txt =~ s/\@l_prefix\@/$prefix/g;
    return $txt;
}

__END__

=pod

=head1 NAME

B<bf-mk> - OpenPKG build farm maker

=head1 SYNOPSIS

The build farm maker initializes the MASTER environment, generates the
SHARED and SLAVES environment, runs the MASTER loop and can query the
config file.

  init mode:      -i [-bu]
  master oneshot: -1
  master loop:    -l
  query mode:     -q
  setup mode:     [-rlkms]

=head1 DESCRIPTION

  init mode

      Option -i invokes init mode. It must be run as the @m_usr@.
      Usually all files are initialized but the list of files can be
      explicitly specified using the -b and -u options.

      -i|init            init mode
      [-b|bfconf]        initialize bf.conf
      [-u|uicgi]         initialize bf-ui.cgi

  master mode

      Option -1 or -l allows the application to execute the MASTER role
      once or in a loop.

      -1|oneshot         run master once
      -l|loop            run master in a infinite loop

  query mode

      Option -q allows printing of variables read from the bf.conf file.

      -q|query=s         query variable from bf.conf

  setup mode

      If neither init nor master nor query mode is used, SHARED setup is
      run. Usually all files are initialized but the list of files can
      be explicitly specified using the -r, -l, -k, -m and -s options.

      [-r|bashrc]        generate .bashrc
      [-p|bashlogin]     generate .bash_login profile
      [-k|key]           generate .ssh/ keys
      [-m|master]        generate bin/master wrapper
      [-s|slave]         generate bin/slave script

  common options

      [-o|outfile=s]     write output to file
      [-C|config=s]      specifiy config file
      [-H|help]          print out this usage page
      [-V|verbose]       enable verbose run-time mode

=head1 EXAMPLES

  openpkg bf-mk -q 'MASTER_PREFIX=%{master.prefix}'

=head1 CONCEPTS

    The buildfarm is made of the following pieces:

    MASTER

        Must be a OpenPKG instance that runs all the administrative
        software including "openpkg bf-mk", "openpkg bf-db" and the
        Apache web server executing a CGI wrapper for "openpkg bf-ui".
        Get the required dependencies by installing "openpkg-tools"
        package "with_bf yes". The master also stores the database.

    DRIVER

        Can be native OS utilities or a OpenPKG instance providing
        common (development) tools, e.g. "make". Must be installed/
        configured on every machine hosting a SLAVE.

    SLAVES

        OpenPKG instances doing the actual build. They are controlled by
        a "slave" shell script which uses the DRIVER's tools.

    SHARED

        A filesystem area accessible by the MASTER and all SLAVES. When
        multiple hosts are involved a NFS share can be used.

    WORKER

        A UNIX shell account using SHARED as it's HOME. It must exist
        on all machines involved in the build farm. This account runs
        the "master" (wrapper for "openpkg bf-mk" on the machine hosting
        MASTER and the (generated) "slave" shell script on every host
        having SLAVES installed.

    Typical setup:

    Install a MASTER, one or more SLAVES with DRIVER instances.

        mkdir /bf
        for i in master slaves driver; do \
            mkdir /d1/$i; \
            ln -s /d1/$i /bf/$i; \
            sh obmtool $i; \
        done

    Create a SHARED directory and a WORKER account

        su - root
        ... /etc/passwd #worker:*:999:999::0:0:Build Farm Worker:/bf/worker:/bf/master/lib/openpkg/bash
        ... /etc/group  #worker:*:999:worker
        mkdir /d1/worker
        ln -s /d1/worker /bf/worker
        chown worker /bf/worker
        chgrp worker /bf/worker

    Initialize MASTER and adjust configuration verbose

        su - master
        openpkg bf-mk --init -V
        vi /bf/master/etc/openpkg-tools/bf.conf

    Allow the WORKER to write the database via "openpkg bf-db"

        su - root
        mkdir        /bf/master/var/openpkg-tools/bf
        chown master /bf/master/var/openpkg-tools/bf
        chgrp worker /bf/master/var/openpkg-tools/bf
        chmod 775    /bf/master/var/openpkg-tools/bf

    Setup DRIVER

        su - root
        /bf/driver/sbin/visudo
        # User privilege specification
        worker  ALL=(ALL) NOPASSWD: ALL

    let WORKER build SLAVES environment on SHARED directory from MASTER config

        su - worker /bf/master/bin/openpkg bf-mk

    enter WORKER area (and optionally repeat previous step)

        su - worker
        master

    run MASTER once

        su - worker
        master -1

    run SLAVE once

        su - worker
        slave -1

    access database

        su - root
        vi /bf/master/etc/apache/apache.conf
        adjust Apache config to listen on favorite <address>:<port>
        /bf/master/bin/openpkg rc apache start
        mozilla http://127.0.0.1/openpkg-cgi/openpkg-tools/bf-ui.cgi

=head1 PACKAGES

    Here is a list of required packages in obmtool.conf format which is
    easy to read for humans and can be used for automation verbatim.

##
##  obmtool.conf -- OpenPKG poor man's Boot, Build & Management Tool Configuration
##  find obmtool at http://www.zfos.org/
##

%common
    @hasfeature TAGFMT || exit 1 # require obmtool to support TAGFMT

%master
    #openpkg build -Zu -D perl-dbi::with_dbd_sqlite perl-dbi perl-www apache rsync cfg shtool openssh | sh build2obmtool.sh
    echo "---- boot/build ${NODE} %${CMD} ----"
    PREFIX=/bf/${CMD}
    @install openpkg-2.2.2-2.2.2 \
        --tag="${CMD}" \
        --prefix="${PREFIX}" \
        --user="${CMD}" --group="${CMD}"
    @install make-3.80-2.2.0
    @install binutils-2.14-2.2.0
    @install gcc-3.4.2-2.2.0
    @install perl-5.8.5-2.2.1
    @install cfg-0.9.7-20041219 --with=perl
    @install sqlite-2.8.15-2.2.0 --with=v3
    @install perl-openpkg-5.8.5-2.2.0
    @install perl-ds-5.8.5-2.2.0
    @install perl-time-5.8.5-2.2.0
    @install readline-5.0.0-2.2.0
    @install perl-term-5.8.5-2.2.0
    @install perl-sys-5.8.5-2.2.0
    @install openssl-0.9.7d-2.2.0
    @install perl-crypto-5.8.5-2.2.0
    @install perl-net-5.8.5-2.2.0
    @install perl-dbi-5.8.5-2.2.0 --with=dbd_sqlite
    @install openpkg-tools-0.8.28-20050124 --with=bf
    @install sharutils-4.3.77-2.2.0
    @install ncurses-5.4.20041009-2.2.0
    @install texinfo-4.7-2.2.0
    @install diffutils-2.8.7-2.2.0
    @install pcre-5.0-2.2.0
    @install grep-2.5.1-2.2.0
    @install bzip2-1.0.2-2.2.0
    @install perl-util-5.8.5-2.2.0
    @install gzip-1.3.5-2.2.0
    @install perl-mail-5.8.5-2.2.0
    @install perl-ssl-5.8.5-2.2.0
    @install perl-parse-5.8.5-2.2.0
    @install zlib-1.2.1-2.2.0
    @install expat-1.95.8-2.2.0
    @install perl-xml-5.8.5-2.2.0
    @install perl-www-5.8.5-2.2.0
    @install gdbm-1.8.3-2.2.0
    @install apache-1.3.31-2.2.3
    @install rsync-2.6.3-2.2.0
    @install shtool-2.0.1-2.2.0
    @install fsl-1.5.0-2.2.0
    @install openssh-3.9p1-2.2.0

%slaves
    echo "---- boot/build ${NODE} %${CMD} ----"
    PREFIX=/bf/${CMD}
    @install openpkg-2.2.2-2.2.2 \
        --tag="${CMD}" \
        --prefix="${PREFIX}" \
        --user="${CMD}" --group="${CMD}"

%driver
    # openpkg build -Zu make binutils gcc sudo shtool openssh | sh build2obmtool.sh
    echo "---- boot/build ${NODE} %${CMD} ----"
    PREFIX=/bf/${CMD}
    @install openpkg-2.2.2-2.2.2 \
        --tag="${CMD}" \
        --prefix="${PREFIX}" \
        --user="${CMD}" --group="${CMD}"
    @install openpkg-2.2.2-2.2.2
    @install make-3.80-2.2.0
    @install binutils-2.14-2.2.0
    @install gcc-3.4.2-2.2.0
    @install fsl-1.5.0-2.2.0
    @install sudo-1.6.8p1-2.2.2
    @install perl-5.8.5-2.2.1
    @install shtool-2.0.1-2.2.0
    @install openssl-0.9.7d-2.2.0
    @install zlib-1.2.1-2.2.0
    @install openssh-3.9p1-2.2.0

=cut

