##
##  OpenPKG bf-ui.pl - OpenPKG build farm user interface (used as CGI)
##  Copyright (c) 2002-2005 The OpenPKG Project <http://www.openpkg.org/>
##  Copyright (c) 2002-2005 Ralf S. Engelschall <rse@engelschall.com>
##  Copyright (c) 2002-2005 Cable & Wireless <http://www.cw.com/>
##
##  Permission to use, copy, modify, and distribute this software for
##  any purpose with or without fee is hereby granted, provided that
##  the above copyright notice and this permission notice appear in all
##  copies.
##
##  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
##  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
##  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
##  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
##  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
##  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
##  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
##  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
##  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
##  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
##  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
##  SUCH DAMAGE.
##

require 5.008;
use strict;
use OpenPKG::Ctx;
use Data::Dumper;
use IO;
use IO::File;
use Getopt::Long;
use OSSP::cfg;
use CGI;
use String::Divert;
use DBI;
use DBD::SQLite;

#   configure optional debugging
$Data::Dumper::Purity = 1;
$Data::Dumper::Indent = 1;
$Data::Dumper::Terse  = 1;

#   catch and hold time
my $time = time();

#   determine OpenPKG information
my $ctx = new OpenPKG::Ctx;
my $progname = $ctx->cmdprog() . " " . $ctx->cmdname();
my $prefix = $ctx->prefix();

#   options with defaults
my $opt_config     = "$prefix/etc/openpkg-tools/bf.conf";
my $opt_help       = 0;
my $opt_verbose    = 0;

#   configuration
my $opt_verbose = undef;
my $opt_help       = 0;
my $opt_verbose    = 0;

sub verbose () {
    my ($msg) = @_;
    printf STDERR "$progname:DEBUG: %s\n", $msg if($opt_verbose);
}

#   this code reading config is shared between bf-mk.pl, bf-db.pl and
#   bf-ui.pl and a candidate to be moved into a perl module shared
#   between the three as soon as such a module is created.
#
sub subst ($$)
{
    my ($cfg, $txt) = @_;
    while ($txt =~ m|%{([^}]+)\.([^}]+)}|) {
        my ($section, $k) = ($1, $2);
        my $v = $cfg->{$section}->{$k};
        if (defined $v) {
            $txt =~ s|%{$section.$k}|$v|g;
        }
        else {
            die "undefined variable %{$section.$k}";
        }
    }
    return $txt;
}

sub readconfig($)
{
    my ($cfgfile) = @_;
    my $cfg = {};
    my $osspcfg = new OSSP::cfg::simple;

    die "no config file specified" unless (defined $cfgfile);
    my $io = new IO::File "<$cfgfile";
    die "cannot open \"$cfgfile\" for reading ($!)" if(not defined $io);
    my $txt = ''; $txt .= $_ while (<$io>);
    $io->close();

    $osspcfg->parse($txt) || die "error parsing config file";
    my $tree = $osspcfg->unpack(-index => '.*') || die "error unpacking config file";

    foreach my $section (qw(master shared driver slaves)) {
        foreach my $k (keys %{$tree->{$section}->[1]}) {
            my $v = $tree->{$section}->[1]->{$k}->[1];
            $v =~ s|%{([^}\.]+)}|%{$section.\1}|g; #expand shortcuts with section omitted
            $cfg->{$section}->{$k} = $v;
            &verbose("section=$section, k=$k, v=$v");
        }
    }
    foreach my $section (keys %{$cfg}) {
        foreach my $k (keys %{$cfg->{$section}}) {
            $cfg->{$section}->{$k} = &subst($cfg, $cfg->{$section}->{$k});
        }
    }
    return($cfg);
}

sub usage ()
{
    my ($level) = @_;
    print "$progname:$level: Usage: $progname [options]\n" .
          "Available options:\n" .
          " -C|config=s        specifiy config file ($opt_config)\n" .
          " -H|help            print out this usage page\n" .
          " -V|verbose         enable verbose run-time mode\n";
    exit(1) if (defined $level);
    exit(0);
}

#   command line parsing
my $result = GetOptions(
    'C|config=s'     => \$opt_config,
    'H|help'         => \$opt_help,
    'V|verbose'      => \$opt_verbose
) || &usage("ERROR");
&usage() if ($opt_help);

#   handle config
my $cfg = &readconfig($opt_config);
my $shared      = $cfg->{"shared"}->{"prefix"}      || die "mandatory setting shared.prefix undefined";
my $dbfile      = $cfg->{"master"}->{"dbfile"}      || die "mandatory setting master.dbfile undefined";
my $lockretries = $cfg->{"master"}->{"lockretries"} || 1;

#   create objects
my $cgi  = new CGI;
my $dbh = &dbinit($dbfile);
my $html = new String::Divert;
$html->overload(1);

##
##  Generate Canvas
##

#   generate outer canvas
$html .= "<table class=\"bf\">\n";
$html .= "  "; $html->fold("bf");
$html .= "</table>\n";

##
##  Generate View Part
##

$html >> "bf";
if ($cgi->param("page") eq "css") {
    $html .= &cssexit();
}
elsif ($cgi->param("page") eq "processing") {
    $html .= &viewmain();
    $html .= &viewprocessingform();
    $html .= &viewprocessing();
}
elsif ($cgi->param("page") eq "status") {
    $html .= &viewmain();
    $html .= &viewstatusform();
    $html .= &viewstatus();
}
elsif ($cgi->param("page") eq "matrix") {
    $html .= &viewmain();
    $html .= &viewmatrix();
}
elsif ($cgi->param("page") eq "log") {
    $html .= &viewmain();
    $html .= &viewlog($cgi->param("file"), $cgi->param("mode"));
}
else {
    $cgi->delete_all();
    $html .= "<h2><a href=\"" . $cgi->url() . "?page=processing\">processing</a></h2>\n";
    $html .= &viewprocessingform();
    $html .= "shows build farm package processing in real time\n";

    $html .= "<h2><a href=\"" . $cgi->url() . "?page=status\">status</a></h2>\n";
    $html .= &viewstatusform();
    $html .= "shows package status per inst (host-arch-os) or package (name-version-release)\n";

    $html .= "<h2><a href=\"" . $cgi->url() . "?page=matrix\">matrix</a></h2>\n";
    $html .= "shows grand total package status\n";
    $html .= "<br>NOTE: takes approximately 100 seconds to compute before data transfer starts\n";
}
$html << 1;

##
##  Generate HTTP Reponse
##

#   HTML skeleton
my $skel = "";
$skel .= "<html>\n";
$skel .= "  <head>\n";
$skel .= "    <link rel=\"stylesheet\" type=\"text/css\" href=\"bfui.cgi?page=css\">\n";
$skel .= "    <style type=\"text/css\">\n";
$skel .= "      html {\n";
$skel .= "        background: #cccccc;\n";
$skel .= "      }\n";
$skel .= "      h1 {\n";
$skel .= "        font-family: sans-serif, helvetica, arial;\n";
$skel .= "        font-weight: bold;\n";
$skel .= "        font-size:   300%;\n";
$skel .= "      }\n";
$skel .= "    </style>\n";
$skel .= "  </head>\n";
$skel .= "  <body>\n";
$skel .= "    <h1>OpenPKG build farm</h1>\n";
$skel .= "    %BODY%\n";
$skel .= "  </body>\n";
$skel .= "</html>\n";

#   undivert HTML output and wrap with skeleton
$html->undivert(0);
$skel =~ s|\%BODY\%|$html|s;
$html = $skel;

#   create HTTP response
my $header = {};
$header->{type} = 'text/html';
$header->{expires} = '+10s';
$header->{refresh} = $cgi->param("vp_autorefresh") . "; " . $cgi->url() . "?page=processing&vp_autorefresh=" . $cgi->param("vp_autorefresh") if ($cgi->param("vp_autorefresh") > 0);
my $http = $cgi->header($header);
$http .= $html;
print STDOUT $http;

#   die gracefully ;-)
$dbh->disconnect();
undef $dbh;
exit(0);

sub cssexit () {
    my $css = '';

    $css .= "/*\n";
    $css .= "**  bfui.cgi - OpenPKG build farm user interface CGI\n";
    $css .= "*/\n";
    $css .= "\n";
    $css .= "TABLE.bf {\n";
    $css .= "    background:      #f0f0f0;\n";
    $css .= "    font-family:     sans-serif, helvetica, arial;\n";
    $css .= "}\n";
    $css .= "\n";
    $css .= ".bf TD.browse {\n";
    $css .= "    width:           100%;\n";
    $css .= "    background:      #d0d0d0;\n";
    $css .= "}\n";
    $css .= ".bf TD.browse SPAN.title {\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       200%;\n";
    $css .= "    color:           #000000;\n";
    $css .= "}\n";
    $css .= ".bf TD.query {\n";
    $css .= "    width:           100%;\n";
    $css .= "    background:      #d0d0d0;\n";
    $css .= "}\n";
    $css .= ".bf TD.query SPAN.title {\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       200%;\n";
    $css .= "    color:           #000000;\n";
    $css .= "}\n";
    $css .= ".bf TD.view {\n";
    $css .= "    width:           100%;\n";
    $css .= "    background:      #d0d0d0;\n";
    $css .= "}\n";
    $css .= ".bf TD.view SPAN.title {\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       200%;\n";
    $css .= "    color:           #000000;\n";
    $css .= "}\n";
    $css .= ".bf TD.result {\n";
    $css .= "    width:           100%;\n";
    $css .= "    background:      #d0d0d0;\n";
    $css .= "}\n";
    $css .= ".bf TD.result SPAN.title {\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       200%;\n";
    $css .= "    color:           #000000;\n";
    $css .= "}\n";

    #   create HTTP response
    my $header = {};
    $header->{type} = 'text/html';
    $header->{expires} = '+10s';
    my $http = $cgi->header($header);
    $http .= $css;
    print STDOUT $http;
    exit(0);
}

sub viewmain()
{
    my $rv;
    my $html;

    $html .= $cgi->start_form();
    $html .= $cgi->submit('page','main');
    $html .= $cgi->end_form;
    return $html;
}

sub viewprocessingform()
{
    my $rv;
    my $html;

    $html .= $cgi->start_form();
    $html .= $cgi->submit('submit','processing');
    $html .= "&nbsp;autorefresh&nbsp;";
    $html .= $cgi->popup_menu(
        -name     => "vp_autorefresh",
        -value    => [ "0", "3", "10", "60", "300", "900", "3600" ],
        -labels   => { "0" => "never", "3" => "3 secs", "10" => "10 secs", "60" => "1 min", "300" => "5 min", "900" => "15 min", "3600" => "1h" },
        -default  => "0",
    ) . "\n";
    $html .= $cgi->hidden(-name=>'page', -value => "processing");
    $html .= $cgi->end_form;
    return $html;
}

sub viewstatusform()
{
    my $rv;
    my $html;

    my @color;
    $rv = $dbh->selectall_arrayref("SELECT DISTINCT st_build_color " .
                                   "FROM status " .
                                   "ORDER BY st_build_color;");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }
    foreach my $i (@{$rv}) {
        push @color, @{$i}->[0];
    }

    my @inst;
    $rv = $dbh->selectall_arrayref("SELECT DISTINCT st_inst_host, st_inst_arch, st_inst_os " .
                                   "FROM status " .
                                   "ORDER BY st_inst_host, st_inst_arch, st_inst_os;");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }
    foreach my $i (@{$rv}) {
        push @inst, @{$i}->[0] . "-" . @{$i}->[1] . "-" . @{$i}->[2];
    }

    $html .= $cgi->start_form();
    $html .= $cgi->submit('submit','host-arch-os');
    $html .= "&nbsp;";
    $html .= $cgi->popup_menu(
        -name     => "vs_inst",
        -value    => \@inst,
    ) . "\n";
    $html .= $cgi->popup_menu(
        -name     => "vs_color",
        -value    => \@color,
        -default  => "",
    ) . "\n";
    $html .= $cgi->hidden(-name=>'page', -value => "status");
    $html .= $cgi->end_form;

    my @package;
    $rv = $dbh->selectall_arrayref("SELECT DISTINCT st_package_name, st_package_version, st_package_release " .
                                   "FROM status " .
                                   "ORDER BY st_package_name, st_package_version, st_package_release;");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }
    foreach my $i (@{$rv}) {
        push @package, @{$i}->[0] . "-" . @{$i}->[1] . "-" . @{$i}->[2];
    }

    $html .= $cgi->start_form();
    $html .= $cgi->submit('submit','name-version-release');
    $html .= "&nbsp;";
    $html .= $cgi->popup_menu(
        -name     => "vs_package",
        -value    => \@package,
    ) . "\n";
    $html .= $cgi->popup_menu(
        -name     => "vs_color",
        -value    => \@color,
        -default  => "",
    ) . "\n";
    $html .= $cgi->hidden(-name=>'page', -value => "status");
    $html .= $cgi->end_form;
    return $html;
}

sub viewlog()
{
    my ($logfile, $mode) = @_;

    my $html;
    if ($logfile =~ m|\.\.| or $logfile =~ m |^/|) {
        die "PANIC:$0: Intruder detected";
    }

    if ( -f "$shared/log/$logfile" and -r "$shared/log/$logfile" and -s "$shared/log/$logfile" ) {
        my $file;
        my $io = new IO::File "<$shared/log/$logfile" or die "unable to read logfile \"$logfile\"";
        $file .= $_ while (<$io>);
        $io->close();
        $file =~ s|&|&amp;|sg;
        $file =~ s|<|&lt;|sg;
        $file =~ s|>|&gt;|sg;
        $file =~ s;^(Executing\((%|--).+?\):.*?)$;<b>$1</b>;mg;
        $file =~ s|^(\+[^\n]*)$|<font color="#333399">$1</font>|mg;
        $file =~ s;^(.*warn.*)$;<font color="#660000"><b>$1</b></font>;mgi;
        $file =~ s;^(.*(error|abort|fail|fatal|bad|illegal|invalid|usage|denied).*)$;<font color="#cc3333"><b>$1</b></font>;mgi;
        $html .= "<pre>\n$file\n</pre>\n";
    }
    else {
        $html .= "Sorry, the logfile \"$logfile\" does not exist yet or not exist any longer";
    }
    return $html;
}

sub viewprocessing()
{
    my $rv;
    my $html;

    $rv = $dbh->selectall_hashref("SELECT * FROM processing;", "pr_inst_host");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }

    $html .= "<h2><a href=\"" . $cgi->url() . "?page=processing\">processing</a></h2>\n";
    $html .= "<table class=\"bf\">\n";
    $html .= "<tr>";
    $html .= "<td><b>&nbsp;</b></td>";
    $html .= "<td><b>host-arch-os</b></td>";
    $html .= "<td><b>name-version-release</b></td>";
    $html .= "<td><b>building&nbsp;since</b></td>";
    $html .= "<td><b>duration</b></td>";
    $html .= "<tr>\n";
    for my $i (sort keys %{$rv}) {
         $html .= "<tr>";

         # color (unused in realtime processing)
         $html .= "<td><img src=\"http://www.openpkg.org/status.img/icon-dia.gif\" alt=\"processing\"></td>";

         # host-arch-os
         my ($host, $arch, $os, $tag) = (%{$rv}->{$i}->{pr_inst_host}, %{$rv}->{$i}->{pr_inst_arch}, %{$rv}->{$i}->{pr_inst_os}, %{$rv}->{$i}->{pr_inst_tag});
         my $inst .= $host . "-" . $arch . "-" . $os;
         $html .= "<td><a href=\"" . $cgi->url() . "?page=status&submit=host-arch-os&vs_inst=$host-$arch-$os\">$inst</a>" . "&nbsp;</td>";

         # name-version-release or "not building"
         my ($name, $version, $release) = (%{$rv}->{$i}->{pr_package_name}, %{$rv}->{$i}->{pr_package_version}, %{$rv}->{$i}->{pr_package_release});
         my $package .= $name . "-" . $version . "-" . $release;
         if ($name eq "") {
             $html .= "<td>not&nbsp;building</td>";
         }
         else {
            my $logfile = "$inst/$package.log.$host.$arch-$os-$tag";
            if ( -f "$shared/log/$logfile" and -r "$shared/log/$logfile" and -s "$shared/log/$logfile" ) {
                $html .= "<td><a href=\"" . $cgi->url() . "?page=log&mode=tail&file=$logfile\">$package</a></td>";
            }
            else {
                $html .= "<td>$package</td>";
            }
         }

         # since or "idle"; duration
         my $heartbeat = %{$rv}->{$i}->{pr_heartbeat};
         if ($heartbeat eq "") {
             $html .= "<td>idle</td>";
             $html .= "<td>&nbsp;</td>";
         }
         else {
            my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($heartbeat);
            $html .= sprintf "<td>%04s-%02s-%02s&nbsp;%02s:%02s:%02s</td>", $year+1900, $mon+1, $mday, $hour, $min, $sec;
            $html .= "<td>";
            my $duration = $time - $heartbeat;
            if ($duration < 60) {
                $html .= $duration . "&nbsp;seconds";
            }
            elsif ($duration < 60 * 60) {
                $html .= int($duration/60) . "&nbsp;minutes";
            }
            elsif ($duration < 60 * 60 * 24) {
                $html .= "<b>" . int($duration/60/60) . "&nbsp;hours</b>";
            }
            elsif ($duration < 60 * 60 * 24 * 7) {
                $html .= "<b>" . int($duration/60/60/24) . "&nbsp;days</b>";
            }
            else {
                $html .= "<b>weeks</b>";
            }
            $html .= "</td>";
         }
         $html .= "</tr>\n";
    }
    $html .= "</table>";
    return $html;
}

#   split inst "bsd1-ix86-freebsd5.3-openpkg" into host="bsd1", arch="ix86", os="freebsd5.3" and tag="openpkg"
sub splitinst ($)
{
    my ($inst) = @_;
    my ($host, $arch, $os, $tag);
    if ($inst =~ m|^(.+)-([^-]+)-([^-]+)-([^-]+)$|) {
        ($host, $arch, $os, $tag) = ($1, $2, $3, $4);
    };
    return ($host, $arch, $os, $tag);
}

#   split package "perl-openpkg-5.8.6-20041203" into name="perl-openpkg", version="5.8.6" and release="20041203"
sub splitpackage ($)
{
    my ($package) = @_;
    my ($name, $version, $release);
    if ($package =~ m|^(.+)-([^-]+)-([^-]+)$|) {
        ($name, $version, $release) = ($1, $2, $3);
    };
    return ($name, $version, $release);
}

sub viewmatrix()
{
    my $html;
    my ($sth, $rv);

    $html .= "<h2><a href=\"" . $cgi->url() . "?page=matrix\">matrix</a></h2>\n";

    my @insts;
    $rv = $dbh->selectall_arrayref("SELECT DISTINCT st_inst_host, st_inst_arch, st_inst_os " .
                                   "FROM status " .
                                   "ORDER BY st_inst_host, st_inst_arch, st_inst_os;");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }
    foreach my $i (@{$rv}) {
        push @insts, @{$i}->[0] . "-" . @{$i}->[1] . "-" . @{$i}->[2]; }

    my @packages;
    $rv = $dbh->selectall_arrayref("SELECT DISTINCT st_package_name, st_package_version, st_package_release " .
                                   "FROM status " .
                                   "ORDER BY st_package_name, st_package_version, st_package_release;");
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }
    foreach my $i (@{$rv}) {
        push @packages, @{$i}->[0] . "-" . @{$i}->[1] . "-" . @{$i}->[2];
    }

    $html .= "<pre>\n";
    $sth = $dbh->prepare(
        "SELECT st_inst_host, st_inst_arch, st_inst_os, st_build_color FROM status WHERE ( " .
        "    st_package_name like ? and st_package_version like ? and st_package_release like ? and " .
        "    st_build_color like ? );"
    );
    foreach my $package (@packages) {
        my ($name, $version, $release) = &splitpackage($package);
        $rv = $sth->execute(
            $name, $version, $release,
            "%"
        );
        if (not $rv) {
            die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
        }
        $rv = $sth->fetchall_arrayref();
        my $row = {};
        for my $i (@{$rv}) {
            $row->{@{$i}->[0] . "-" . @{$i}->[1] . "-" . @{$i}->[2]} = @{$i}->[3];
        }
        foreach my $inst (@insts) {
            my ($host, $arch, $os, $tag) = &splitinst($inst . "-openpkg"); # bsd1-ix86-freebsd5.3-openpkg

            my $logfile = "$inst/$package.log.$host.$arch-$os-$tag";
            my ($l, $r);
            if ( -f "$shared/log/$logfile" and -r "$shared/log/$logfile" and -s "$shared/log/$logfile" ) {
                $l = "<a href=\"" . $cgi->url() . "?page=log&mode=tail&file=$logfile\">";
                $r = "</a>";
            }
            else {
                $l = "";
                $r = "";
            }

            my $color = $row->{$inst};
            if ($color eq "green") {
                $html .= "$l<img src=\"http://www.openpkg.org/status.img/icon-ok.gif\" alt=\"green\">$r";
            }
            elsif ($color eq "red") {
                $html .= "$l<img src=\"http://www.openpkg.org/status.img/icon-x.gif\" alt=\"red\">$r";
            }
            else {
                $html .= "$l<img src=\"http://www.openpkg.org/status.img/icon-dot.gif\" alt=\"$color\">$r";
            }
        }
        $html .= "&nbsp;<a href=\"" . $cgi->url() . "?page=status&submit=name-version-release&vs_package=$package\">$package</a>\n";
    }
    $html .= "</pre>";
    return $html;
}

sub viewstatus()
{
    my ($host, $arch, $os, $tag)   = &splitinst("%-%-%-%");  # bsd1-ix86-freebsd5.3-openpkg
    my ($name, $version, $release) = &splitpackage("%-%-%"); # perl-5.8.6-20050111
    my ($color) = "%";                                       # red
    my ($sth, $rv);
    my $html;

    $html .= "<h2><a href=\"" . $cgi->url() . "?page=status\">status</a></h2>\n";

    if ($cgi->param("submit") eq "host-arch-os") {
        ($host, $arch, $os, $tag) = &splitinst($cgi->param("vs_inst") . "-openpkg");
        ($name, $version, $release) = &splitpackage("%-%-%")
    }
    if ($cgi->param("submit") eq "name-version-release") {
        ($host, $arch, $os, $tag) = &splitinst("%-%-%-openpkg");
        ($name, $version, $release) = &splitpackage($cgi->param("vs_package"));
    }
    if ($cgi->param("vs_color") eq "") {
        $color = "%";
    }
    else {
        $color = $cgi->param("vs_color");
    }

    $sth = $dbh->prepare(
        "SELECT * FROM status WHERE ( " .
        "    st_inst_host like ? and st_inst_arch like ? and st_inst_os like ? and st_inst_tag like ? and " .
        "    st_package_name like ? and st_package_version like ? and st_package_release like ?  and " .
        "    st_build_color like ? );"
    );
    $rv = $sth->execute(
        $host, $arch, $os, $tag,
        $name, $version, $release,
        $color
    );
    if (not $rv) {
        die "ERROR:$0: SQLite error: ".$dbh->errstr."\n";
    }
    $rv = $sth->fetchall_arrayref({});

    $html .= "<table class=\"bf\">\n";
    $html .= "<tr>";
    $html .= "<td><b>&nbsp;</b></td>";
    $html .= "<td><b>host-arch-os</b></td>";
    $html .= "<td><b>time</b></td>";
    $html .= "<td><b>spec</b></td>";
    $html .= "<td><b>dir</b></td>";
    $html .= "<td><b>name-version-release</b></td>";
    $html .= "<tr>\n";
    for my $i (@{$rv}) {
         $html .= "<tr>";

         my $color = %{$i}->{st_build_color};
         my ($host, $arch, $os, $tag) = (%{$i}->{st_inst_host}, %{$i}->{st_inst_arch}, %{$i}->{st_inst_os}, %{$i}->{st_inst_tag});
         my ($name, $version, $release) = (%{$i}->{st_package_name}, %{$i}->{st_package_version}, %{$i}->{st_package_release});

         # color
         if ($color eq "green") {
             $html .= "<td><img src=\"http://www.openpkg.org/status.img/icon-ok.gif\" alt=\"green\"></td>";
         }
         elsif ($color eq "red") {
             $html .= "<td><img src=\"http://www.openpkg.org/status.img/icon-x.gif\" alt=\"red\"></td>";
         }
         else {
             $html .= "<td><img src=\"http://www.openpkg.org/status.img/icon-dot.gif\" alt=\"$color\"></td>";
         }

         # host-arch-os
         my $inst .= $host . "-" . $arch . "-" . $os;
         $html .= "<td><a href=\"" . $cgi->url() . "?page=status&submit=host-arch-os&vs_inst=$host-$arch-$os\">$inst</a>" . "&nbsp;</td>";

         # cvs
         $html .= "<td><a href=\"http://cvs.openpkg.org/rlog?f=openpkg-src/$name/$name.spec\">time</a></td>";
         $html .= "<td><a href=\"http://cvs.openpkg.org/getfile/openpkg-src/$name/$name.spec\">spec</a></td>";
         $html .= "<td><a href=\"http://cvs.openpkg.org/dir?d=openpkg-src/$name\">dir</a></td>";

         # name-version-release or "not building"
         my $package .= $name . "-" . $version . "-" . $release;
         my $logfile = "$inst/$package.log.$host.$arch-$os-$tag";
         if ( -f "$shared/log/$logfile" ) {
             $html .= "<td>";
             $html .= "<a href=\"" . $cgi->url() . "?page=status&submit=name-version-release&vs_package=$name-$version-$release\">$name</a>" . "&nbsp;";
             $html .= "-$version-";
             $html .= "<a href=\"" . $cgi->url() . "?page=log&mode=highlight&file=$logfile\">$release</a>";
             $html .= "</td>";
         }
         else {
             $html .= "<td>";
             $html .= "<a href=\"" . $cgi->url() . "?page=status&submit=name-version-release&vs_package=$name-$version-$release\">$name</a>" . "&nbsp;";
             $html .= "-$version-";
             $html .= "$release";
             $html .= "</td>";
         }
         $html .= "</tr>\n";
    }
    $html .= "</table>";
    return $html;
}

#   open database and check whether important tables exist
sub dbinit ($)
{
    my ($dbfile) = @_;
    my $dbh = DBI->connect("dbi:SQLite:dbname=$dbfile", "", "", { RaiseError => 0, AutoCommit => 1 });
    if (not defined $dbh) { die "$0: \"cannot open database file \"$dbfile\"\n"; }

	my $tables = $dbh->selectall_hashref("SELECT name FROM sqlite_master WHERE type='table' ORDER BY name;", "name");

	unless ($tables->{status})     { die "$0: \"status\" table missing in database file \"$dbfile\"\n"; }
	unless ($tables->{processing}) { die "$0: \"processing\" table missing in database file \"$dbfile\"\n"; }
	unless ($tables->{class})      { die "$0: \"class\" table missing in database file \"$dbfile\"\n"; }

    return $dbh;
}

#   internal handling of fatal errors
BEGIN {
    $SIG{__DIE__} = sub {
        my ($msg) = @_;
        my $hint = '';
        if ($msg =~ m|line\s+(\d+)|) {
            my $line = $1;
            my $io = new IO::File "<$0";
            my @code = $io->getlines();
            $io->close();
            my $i = -1;
            $hint = join("", map { s/^/sprintf("%d: ", $line+$i++)/se; $_; } @code[$line-2..$line]);
        }
        print STDOUT 
            "Content-Type: text/html; charset=ISO-8859-1\n" .
            "\n" .
            "<html>\n" .
            "  <head>\n" .
            "    <title>GURU MEDITATION</title>\n" .
            "    <style type=\"text/css\">\n" .
            "      HTML {\n" .
            "          width:  100%;\n" .
            "          height: auto;\n" .
            "      }\n" .
            "      BODY {\n" .
            "          background: #cccccc;\n" .
            "          margin:     0 0 0 0;\n" .
            "          padding:    0 0 0 0;\n" .
            "      }\n" .
            "      DIV.canvas {\n" .
            "          background: #000000;\n" .
            "          border: 20px solid #000000;\n" .
            "      }\n" .
            "      DIV.error1 {\n" .
            "          border-top:          4px solid #cc3333;\n" .
            "          border-left:         4px solid #cc3333;\n" .
            "          border-right:        4px solid #cc3333;\n" .
            "          border-bottom:       4px solid #cc3333;\n" .
            "          padding:             10px 10px 10px 10px;\n" .
            "          font-family:         sans-serif, helvetica, arial;\n" .
            "          background:          #000000;\n" .
            "          color:               #cc3333;\n" .
            "      }\n" .
            "      DIV.error2 {\n" .
            "          border-top:          4px solid #000000;\n" .
            "          border-left:         4px solid #000000;\n" .
            "          border-right:        4px solid #000000;\n" .
            "          border-bottom:       4px solid #000000;\n" .
            "          padding:             10px 10px 10px 10px;\n" .
            "          font-family:         sans-serif, helvetica, arial;\n" .
            "          background:          #000000;\n" .
            "          color:               #cc3333;\n" .
            "      }\n" .
            "      SPAN.title {\n" .
            "          font-size:           200%;\n" .
            "          font-weight:         bold;\n" .
            "      }\n" .
            "      TT.text {\n" .
            "          font-weight:         bold;\n" .
            "      }\n" .
            "    </style>\n" .
            "    <script language=\"JavaScript\">\n" .
            "    var count = 0;\n" .
            "    function blinker() {\n" .
            "        var obj = document.getElementById('error');\n" .
            "        if (count++ % 2 == 0) {\n" .
            "            obj.className = 'error1';\n" .
            "        }\n" .
            "        else {\n" .
            "            obj.className = 'error2';\n" .
            "        }\n" .
            "        setTimeout('blinker()', 1000);\n" .
            "    }\n" .
            "    </script>\n" .
            "  </head>\n" .
            "  <body onLoad=\"setTimeout('blinker()', 1);\">\n" .
            "    <div class=\"canvas\">\n" .
            "      <div id=\"error\" class=\"error1\">\n" .
            "        <span class=\"title\">GURU MEDITATION</span>\n" .
            "        <p>\n" .
            "        <tt class=\"text\">\n" .
            "          $msg<br>\n" .
            "        </tt>\n" .
            "        <pre>\n$hint</pre>\n" .
            "      </div>\n" .
            "    </div>\n" .
            "  </body>\n" .
            "</html>\n";
        exit(0);
    };
}

__END__

=pod

=head1 NAME

B<bfui> - OpenPKG build farm user interface

=head1 SYNOPSIS

CGI frontend to build farm database.

=head1 DESCRIPTION

  common options

      [-o|outfile=s]     write output to file
      [-C|config=s]      specifiy config file
      [-H|help]          print out this usage page
      [-V|verbose]       enable verbose run-time mode

Accessible through 
http://127.0.0.1/openpkg-cgi/openpkg-tools/bf-ui.cgi
after master initialization has been run, see "openpkg bf-mk".

=cut

